use crate::{
    entity::prelude::Account,
    errors::NoUserInfo,
    handlers::{markups::deletion_markup, utils::package_handler, MainDialogue, State},
    models::{DecryptedAccount, User},
};
use futures::TryStreamExt;
use sea_orm::DatabaseConnection;
use serde_json::to_vec_pretty;
use std::sync::Arc;
use teloxide::{adaptors::Throttle, prelude::*, types::InputFile};
use tokio::{sync::Mutex, task::spawn_blocking};

async fn get_master_pass(
    bot: Throttle<Bot>,
    msg: Message,
    db: DatabaseConnection,
    dialogue: MainDialogue,
    master_pass: String,
) -> crate::Result<()> {
    let master_pass: Arc<str> = master_pass.into();
    let user_id = msg.from().ok_or(NoUserInfo)?.id.0;
    let accounts = Arc::new(Mutex::new(Vec::new()));
    Account::get_all(user_id, &db)
        .await?
        .try_for_each_concurrent(None, |account| {
            let master_pass = Arc::clone(&master_pass);
            let accounts = Arc::clone(&accounts);
            async move {
                let account =
                    spawn_blocking(move || DecryptedAccount::from_account(account, &master_pass))
                        .await??;
                accounts.lock().await.push(account);
                Ok(())
            }
        })
        .await?;
    drop(master_pass);
    let mut accounts = match Arc::try_unwrap(accounts) {
        Ok(account) => account.into_inner(),
        Err(_) => {
            return Err(crate::Error::msg(
                "Couldn't get accounts from Arc in export.rs",
            ))
        }
    };
    accounts.sort_unstable_by(|this, other| this.name.cmp(&other.name));
    let json = to_vec_pretty(&User { accounts })?;
    let file = InputFile::memory(json).file_name("accounts.json");
    bot.send_document(msg.chat.id, file)
        .reply_markup(deletion_markup())
        .await?;
    dialogue.exit().await?;
    Ok(())
}

pub async fn export(bot: Throttle<Bot>, msg: Message, dialogue: MainDialogue) -> crate::Result<()> {
    let previous = bot
        .send_message(msg.chat.id, "Send a master password to export the accounts")
        .await?;
    dialogue
        .update(State::GetMasterPass(package_handler(
            move |bot, msg, db, dialogue, master_pass| {
                get_master_pass(bot, msg, db, dialogue, master_pass)
            },
            previous,
        )))
        .await?;
    Ok(())
}
